/* ****************************************** */
/* Emu71: HP-71B software emulator for PC     */
/*                                            */
/* hpil.c   HPIL module emulation             */
/*                                            */
/* Version 2.43, J-F Garnier 2003-2012        */
/* ****************************************** */




#include "saturn.h"

#include "version.h"

extern char f_trace;   /* trace */


#define MBRESET 1
#define MBIDLE 0
#define ACT_LISTEN 2
#define END_LISTEN 3


/* ************************************************************* */
/* local variables                                               */
/* note: the management of 3 loops was planned originally,       */
/* so the state array has room for all the loops                 */
/* However, only one loop has been implemented in the code       */
/* ************************************************************* */

static int state[4] = {MBIDLE,MBIDLE,MBIDLE,MBIDLE};

/* loop variables :*/
/* loop status (6 nibbles): BPUI STLC, KRXV 0001, NNNN NNNN */
static int mbstatus[3] = {0x09, 0x01, 0 };
static int mbnbdev=0;    /* number of devices on the loop */
static long maxcount, count;  /* max count and actual count during transfer */
static int lastframe;    /* last frame */
static int buffer[4];    /* last 3 received frames + possible termination frame */
static int lenbuf, ptbuf;/* length and present pointer to the buffer */
static char fendmode, fcharmode; /* transfer end mode, terminator character mode */
static unsigned char endchar;    /* transfer terminator character */
static int currentaddr;          /* current HPIL address for transfer */
static char fmanual;     /* 0=auto, 1=manual, 2=scope  (24/06/12) */

#define hpil_transmit1(n) hpil_transmit(n,1)


/* ****************************************** */
/* init_hpil()                                */
/*                                            */
/* init a loop by sending AAU, AAD 1,         */
/* set mbnbdev: mailbox number of devices     */
/* ****************************************** */
void init_hpil(void)
{
  int n;

  hpil_transmit1(0x49a); /* AAU */
  n=hpil_transmit1(0x581); /* AAD 1 */
  mbnbdev=n-0x581;
}



/* ****************************************** */
/* hpil_ctrl()                                */
/*                                            */
/* check the controller bit                   */
/* ****************************************** */
int hpil_ctrl(void)
{
  return(mbstatus[0]&1);
}


/* ****************************************** */
/* hpil_transmit()                            */
/*                                            */
/* transmit a frame on the loop               */
/* flag f means XIL/XIL2 allowed              */
/* ****************************************** */
int hpil_transmit(int n, char f)
{
  int i;

  for (i=0;i<nbdev;i++) {
    if (n<0) break;
    switch (tabdev[i]) {
      case 1: /* DISPLAY */
        n=ildisplay(n);
        break;
      case 2: /* HDRIVE1 */
        n=ilhdisc(n);
        break;
      case 3: /* FDRIVE1 */
        n=ilfdisc(n);
        break;
      case 4: /* DOSLINK */
       n=ildos(n);
        break;
      case 5: /* XIL */
#ifdef XIL
        if (f) {
  	  if (n==0x49a) ilext(0x600);  /* IDY to check external loop integrity */
          n=ilext(n);
        }
#endif
        break;
      case 6: /* LPRTER1 */
	n=ilcentro(n);
        break;
      case 7: /* HDRIVE2 */
        n=ilhdisc2(n);
        break;
      case 8: /* SERIAL1 */
        n=ilserial1(n);
        break;
      case 9: /* SERIAL2 */
        n=ilserial2(n);
        break;
      case 10: /* XIL2 */
        if (f) {;
  	  if (n==0x49a) ilext2(0x600);  /* IDY to check external loop integrity */
          n=ilext2(n);
        }
        break;
//      case 11: /* PIO1 */
//        n=ilpio1(n);
//        break;

    }
  }
  return(n);
}









/* ****************************************** */
/* send3bytes()                               */
/*                                            */
/* transmit 3 data bytes to the loop          */
/*   mb: mailbox (not used, 0 is used)        */
/*   msg[0..2]: byte array                    */
/* ****************************************** */
void static send3bytes(int mb, unsigned char *msg)
{
  int i, n;
  for (i=0;i<3;i++)
    n=hpil_transmit1(msg[i]);
}

/* ****************************************** */
/* send1byte()                                */
/*                                            */
/* transmit 1 data byte to the loop           */
/*   mb: mailbox (not used, 0 is used)        */
/*   c: character to send                     */
/* ****************************************** */
void static send1byte(int mb, unsigned char c)
{
  int n;

  n=hpil_transmit1(c);
}


/* ****************************************** */
/* enter_bytes()                              */
/*                                            */
/* enter 1 to 3 data bytes from the loop      */
/* manage the end of transfer condition       */
/*   mb: mailbox (not used, 0 is used)        */
/*   startframe: SOT frame to use             */
/* ****************************************** */
int static enter_bytes(int mb, int startframe)
{
  int n, f;

  lenbuf=0;
  n=startframe;
  do {
    n=hpil_transmit1(n);
    lastframe=n;
    if ((startframe&0x500)&&(n==startframe)) return(-1); /* error or no response */
    buffer[lenbuf++]=n;
    if ((n==0x540)||(n==0x541)) return(0);  /*EOT ETE */
    count++;
    f=0;
    if (count>=maxcount) f=1;
    if (fendmode&&(n&0x200)) f=2;
    if (fcharmode&&((n&0xff)==endchar)) f=2;
    if (f) {
      /* send NRD */
      n=hpil_transmit1(0x542);  /* NRD */
      n=hpil_transmit1(lastframe);
      if (f==2) buffer[lenbuf++]=-1;  /* term cond met */
      return(0);
    }
    if (count==1) return(1); /* returns immediatly the 1st byte (needed by INITIALIZE) */
    if (lenbuf==3) return(3);
  } while (1);

}


/* ****************************************** */
/* end_listen()                               */
/*                                            */
/* process the loop when in                   */
/*   end_listen state                         */
/*   mb: mailbox number (0 is assumed)        */
/*   msg: mailbox data                        */
/* ****************************************** */
void static end_listen(int mb, unsigned char *msg)
{
  int n;

  n=buffer[ptbuf++];
  if (n<0) {
    /* term cond met */
    msg[4] = 0;
    msg[5]=0x05;
    msg[6]=0;
    msg[7]=0;
    /* set bit msg available */
    msg[4] |= 1;
    state[mb]=MBIDLE;
  }
  else if ((n==0x540)||(n==0x541)) {
    /* EOT */
    msg[4] = 0;
    msg[5]=0x02;
    msg[6]=0;
    msg[7]=0;
    /* set bit msg available */
    msg[4] |= 1;
    state[mb]=MBIDLE;
  }
  else {
    msg[4] = 0;
    msg[5]=n;
    msg[6]=0x08;
    msg[7]=0;
    /* set bit msg available */
    msg[4] |= 1;
  }
  if (ptbuf==lenbuf) state[mb]=MBIDLE;
}


/* ****************************************** */
/* do_listen()                                */
/*                                            */
/* process the loop when in                   */
/* active listen state                        */
/*   mb: mailbox number (0 is assumed)        */
/*   msg: mailbox data                        */
/*   frame: SOT frame or last received frame  */
/* ****************************************** */
int static do_listen(int mb, unsigned char *msg, int frame)
{
  int n;

  /* enter bytes until EOT or count */
  n=enter_bytes(mb, frame);  /* return: -1:error, 0: bytes+EOT: 1: 3bytes */
  switch (n) {
    case -1:
      /* error */
      /* no response --> device not ready error*/
      mbstatus[2]=2;
      /* set error bit */
      msg[4] = 0x10;
      /* send status */
      msg[5]=mbstatus[0];
      msg[6]=mbstatus[1]|1;
      msg[7]=mbstatus[2];
      /* set bit msg available */
      msg[4] |= 1;
      state[mb]=MBIDLE;
      break;
    case 0:
      /* conversation is over (by count, EOT or cond) */
      ptbuf=0;
      state[mb]=END_LISTEN;
      end_listen(mb, msg);
      break;
    case 1:
      /* received 1 bytes, transmit to 71 and go on */
      msg[4] = 0;
      msg[5]=buffer[0];
      msg[6]=0x08;
      msg[7]=0;
      state[mb]=ACT_LISTEN;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 3:
      /* received 3 bytes, transmit to 71 and go on */
      msg[4] = 0x80;
      msg[5]=buffer[0];
      msg[6]=buffer[1];
      msg[7]=buffer[2];
      state[mb]=ACT_LISTEN;
      /* set bit msg available */
      msg[4] |= 1;
      break;
  }
  return(0);
}


/* ****************************************** */
/* findndevice()                              */
/*                                            */
/* find the nth device of given type          */
/* ****************************************** */
int static findndevice(int type, int nb)
{
  int i, n, cpt;

  n=hpil_transmit1(0x43f); /* UNL */
  cpt=0;
  for (i=1;i<=mbnbdev;i++) {
    n=hpil_transmit1(0x440+i); /* TAD currentaddr */
    n=hpil_transmit1(0x563); /* SAI */
    if (n!=0x563 ) { /* check for response */
      hpil_transmit1(n); /*  return the AID byte to get the EOT - fixed 19/09/12, from C.G. ) */
      if ((type&0x0f)==0x0f) {
        /* test general type */
        if ((type&0xf0)==(n&0xf0)) cpt++;
      }
      else
        if (type==n) cpt++;
      if (cpt==nb) break;
    }
  }
  if (cpt!=nb) {
    n=hpil_transmit1(0x45f); /* UNT */
    i=-1;
  }
  return(i);
}



/* ****************************************** */
/* execmd0()                                  */
/*                                            */
/* execution of no parameter class opcodes    */
/* note: several opcodes are not implemented  */
/* in Emu71, they are indicated with '!' at   */
/* the end of the trace messages              */
/* ****************************************** */
void static execmd0(int mb, unsigned char *msg)
{
  int opcode;

  opcode=msg[2]&15;

  switch (opcode) {
    case 0:  /* nop */
      if (f_trace&2) disp_string("HPIL: nop !\n\r");
      break;
    case 1:  /* read address table */
      if (f_trace&2) disp_string("HPIL: read address table\n\r");
      msg[4]=0;
      msg[5]=mbnbdev;
      msg[6]=0;
      msg[7]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 2:  /* request hpil status */
      if (f_trace&2) disp_string("HPIL: request hpil status\n\r");
      /* send status */
      msg[4]=0;
      msg[5]=mbstatus[0];
      msg[6]=mbstatus[1]|1;
      msg[7]=mbstatus[2];
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 3:  /* end of message */
      if (f_trace&2) disp_string("HPIL: end of message !\n\r");
      break;
    case 4:  /* clear SRQ */
      if (f_trace&2) disp_string("HPIL: clear SRQ !\n\r");
      break;
    case 5:  /* set SRQ */
      if (f_trace&2) disp_string("HPIL: set SRQ !\n\r");
      break;
    case 6:  /* send error msg */
      if (f_trace&2) disp_string("HPIL: send error msg\n\r");
      /* clear error bit */
      msg[4]=0;
      /* send status */
      msg[5]=mbstatus[0];
      msg[6]=mbstatus[1]|1;
      msg[7]=mbstatus[2];
      /* clear error */
      mbstatus[2]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 7:  /* enter auto end mode */
      if (f_trace&2) disp_string("HPIL: enter auto end mode !\n\r");
      break;
    case 8:  /* go into manual mode */
      if (f_trace&2) disp_string("HPIL: go into manual mode !\n\r");
      if (msg[1]&1) fmanual=2; else fmanual=1;  /* 24/06/12 */
      break;
    case 9:  /* go into auto mode */
      if (f_trace&2) disp_string("HPIL: go into auto mode !\n\r");
      fmanual=0;                                /* 24/06/12 */
      break;
    case 10: /* system ctrl bit */
      if (f_trace&2) disp_string("HPIL: system ctrl bit !\n\r");
      break;
    case 11: /* reset current address */
      if (f_trace&2) disp_string("HPIL: reset current address\n\r");
      currentaddr=1;
      break;
    case 12: /* read current address */
      if (f_trace&2) disp_string("HPIL: read current address\n\r");
      msg[4]=0;
      msg[5]=currentaddr;
      msg[6]=0x04;
      msg[7]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 13: /* increment current address */
      if (f_trace&2)  disp_string("HPIL: increment current address\n\r");
      if (currentaddr<mbnbdev) {
        currentaddr++;
        msg[4]=0;
        msg[5]=currentaddr;
        msg[6]=0x04;
        msg[7]=0;
        /* set bit msg available */
        msg[4] |= 1;
      }
      else {
        currentaddr=1;
        mbstatus[2]=12;  /* illegal current adress */
        /* set error bit */
        msg[4] = 0x10;
        /* send status */
        msg[5]=mbstatus[0];
        msg[6]=mbstatus[1]|1;
        msg[7]=mbstatus[2];
        /* set bit msg available */
        msg[4] |= 1;
      }
      break;
    case 14: /* read my loop address */
      if (f_trace&2)  disp_string("HPIL: read my loop address\n\r");
      msg[4]=0;
      msg[5]=0;         /* 0 */
      msg[6]=0x04;
      msg[7]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 15: /* take/give loop control */
      if (f_trace&2) disp_string("HPIL: take/give loop control\n\r");
      /* JFG 09/04/06 :*/
      mbstatus[0] &=0xf0;
      if (msg[1]&1) {
        mbstatus[0] |=1;
#ifdef XIL
        init_ilext(0,1);
        init_ilext2(0,1);  /* added JFG 14/04/08 */
      }
      else {
        init_ilext(0,0);
        init_ilext2(0,0);  /* added JFG 14/04/08 */
#endif
      }
      break;
  }
}

/* ****************************************** */
/* execmd15()                                 */
/*                                            */
/* execution of multibyte parameter class     */
/*   opcodes                                  */
/* ****************************************** */
void static execmd15(int mb, unsigned char *msg)
{
  int opcode, n;

  opcode=msg[2]&15;

  switch (opcode) {
    case 0:  /* read memory */
      break;
    case 1:  /* write memory */
      break;
    case 2:  /* self test */
      break;
    case 3:  /* set SOT reponse */
      if (f_trace&2) disp_string("HPIL: set SOT reponse !\n\r");
      break;
    case 4:  /* set terminator mode */
     if (f_trace&2)  disp_string("HPIL: set terminator mode\n\r");
      n=msg[1];
      if (n&8) {
        if (n&4) fendmode=1; else fendmode=0;
      }
      else {
        if (n&1) fcharmode=1; else fcharmode=0;
      }
      break;
    case 5:  /* set terminator character */
     if (f_trace&2)  disp_string("HPIL: set terminator character\n\r");
      endchar=msg[1];
      break;
    case 6:  /* set number of IDY timeout */
      if (f_trace&2) disp_string("HPIL: set number of IDY timeout !\n\r");
      break;
    case 7:  /* set IDY timeout value*/
      if (f_trace&2) disp_string("HPIL: set IDY timeout value !\n\r");
      break;
    case 8:  /* clear data buffers */
      if (f_trace&2) disp_string("HPIL: clear data buffers !\n\r");
      break;
    case 9:  /* set IDY poll timeout */
      if (f_trace&2) disp_string("HPIL: set IDY poll timeout !\n\r");
      break;
    case 10: /* set up interrupt mask */
      if (f_trace&2) disp_string("HPIL: set up interrupt mask !\n\r");
      break;
    case 11: /* read interrupt cause */
      if (f_trace&2) disp_string("HPIL: read interrupt cause !\n\r");
      msg[4]=0;
      msg[5]=0;  /* 0 */
      msg[6]=3;
      msg[7]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 12: /* read DDC frame */
      if (f_trace&2) disp_string("HPIL: read DDC frame !\n\r");
      msg[4]=0;
      msg[5]=0;  /* 0 */
      msg[6]=3;
      msg[7]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 13: /* terminate on loop SRQ */
      if (f_trace&2) disp_string("HPIL: terminate on loop SRQ !\n\r");
      break;
    case 14: /* power up the loop */
      if (f_trace&2) disp_string("HPIL: power up the loop\n\r");
      hpil_transmit1(0x400); /* NOP */
      /* send status */
      msg[4]=0;
      msg[5]=mbstatus[0];
      msg[6]=mbstatus[1]|1;
      msg[7]=mbstatus[2];
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 15: /* enable/disable IDY poll */
      if (f_trace&2) disp_string("HPIL: enable/disable IDY poll !\n\r");
      break;
  }
}


/* ****************************************** */
/* execmd()                                   */
/*                                            */
/* execution of mailbox opcodes               */
/*   mb: mailbox number (0 is assumed)        */
/*   msg: mailbox data                        */
/* ****************************************** */
void static execmd(int mb, unsigned char *msg)
{
  int opcode;
  int n;

  opcode=msg[2]>>4;

  switch (opcode) {

    case 0:  /* no parameter class */
      execmd0(mb,msg);
      break;

    case 1:  /* frame class */
      /* send a frame */
      if (f_trace&2) disp_string("HPIL: send a frame\n\r");
      if (mbstatus[0]&1) {
        n=((msg[2]<<8)&0x700)+msg[1];
        n=hpil_transmit1(n);
        if (msg[2]&0x08) {
          msg[5]=n&0xff;
          msg[6]=0x10+(n>>8);
          msg[7]=0;
          /* set bit msg available */
          msg[4] |= 1;
        }
      }
      else {
        /* not controller */
        mbstatus[2]=9;
        msg[4] = 0x10;
      }
      break;

    /* single nibble parameter class */
    case 2:  /* (un)address me as TL */
      if (f_trace&2) disp_string("HPIL: (un)address me as TL\n\r");
      n=msg[1]&0x16;
      mbstatus[0] &= 0xf9;
      if (n==0x02) {
        /* MLA */
        mbstatus[0] |= 2;
      }
      else if (n==0x04) {
        /* MTA */
        hpil_transmit1(0x45f);  /* UNT */
        mbstatus[0] |= 4;
      }
      break;
    case 3:  /* power down the loop */
      if (f_trace&2) disp_string("HPIL: power down the loop !\n\r");
      break;
    case 4:  /* address P,S as talker */
      if (f_trace&2) disp_string("HPIL: address P,S as talker\n\r");
      n=msg[1]&0x1f;
      if (n==0) n=currentaddr;
      mbstatus[0] &= ~4;
      hpil_transmit1(0x440+n);  /* TAD n */
      break;
    case 5:  /* address P,S as listener */
      if (f_trace&2) disp_string("HPIL: address P,S as listener\n\r");
      n=msg[1]&0x1f;
      if (n==0) n=currentaddr;
      if (n==31) mbstatus[0] &= ~2;
      hpil_transmit1(0x420+n);  /* LAD n */
      break;
    case 6:  /* find Nth device, type M */
     if (f_trace&2)  disp_string("HPIL: find Nth device, type M\n\r");
      n=findndevice(msg[1],(msg[2]&15)+1);
      if (n>0) {
        currentaddr=n;
        msg[4]=0;
        msg[5]=currentaddr;
        msg[6]=0x04;
        msg[7]=0;
        /* set bit msg available */
        msg[4] |= 1;
      }
      else {
        currentaddr=1;
        mbstatus[2]=1;  /* no such device */
        /* set error bit */
        msg[4] = 0x10;
        /* send status */
        msg[5]=mbstatus[0];
        msg[6]=mbstatus[1]|1;
        msg[7]=mbstatus[2];
        /* set bit msg available */
        msg[4] |= 1;
      }
      break;
    case 7:  /* auto address the loop */
      if (f_trace&2) disp_string("HPIL: auto address the loop\n\r");
      hpil_transmit1(0x49a); /* AAU */
      n=hpil_transmit1(0x581); /* AAD 1 */
      mbnbdev=n-0x581;
      /* send nb of devices */
      msg[5]=mbnbdev;
      msg[6]=0x04;
      msg[7]=0;
      /* set bit msg available */
      msg[4] |= 1;
      break;

    /* conversation class */
    case 8:  /* start data */
    case 9:  /* start status poll */
    case 10: /* start device id */
    case 11: /* start accessory id */
      if (f_trace&2) disp_string("HPIL: start conversation\n\r");
      maxcount=msg[0]+msg[1]*256L+(msg[2]&15)*65536L;
      count=0;
      n=0x560+(opcode-8);  /* SDA to SAI */
      do_listen(mb, msg, n);
      break;
    case 12: /* pass control */
      if (f_trace&2) disp_string("HPIL: pass control !\n\r");
      mbstatus[2]=2;      /* device not ready error*/
      /* set error bit */
      msg[4] = 0x10;
      /* send status */
      msg[5]=mbstatus[0];
      msg[6]=mbstatus[1]|1;
      msg[7]=mbstatus[2];
      /* set bit msg available */
      msg[4] |= 1;
      break;
    case 13: /* set timeout value */
      if (f_trace&2) disp_string("HPIL: set timeout value !\n\r");
      break;
    case 14: /* set frame count */
      if (f_trace&2) disp_string("HPIL: set frame count !\n\r");
      break;

    case 15: /* multibyte parameter class*/
      execmd15(mb,msg);
      break;

  }
}


/* ****************************************** */
/* reset_mb()                                 */
/*                                            */
/* reset a HPIL mailbox                       */
/*   mb: mailbox number (0..2)                */
/* ****************************************** */
int reset_mb(int mb)
{

  if (f_trace&2) disp_string("HPIL: mailbox reset\n\r");

  /* following is only for loop 1: */
  if (mb==0) {
    mbstatus[0] = 0x09;
    mbstatus[1] = 0x01;
    mbstatus[2] = 0;
    maxcount=0; count=0;
    lastframe=-1;
    lenbuf=0; ptbuf=0;
    fendmode=0; fcharmode=0;
    currentaddr=0;
    fmanual=0;           /* 24/06/12 */
    init_hpil();
  }
}


/* ****************************************** */
/* ilmailbox()                                */
/*                                            */
/* process the hpil mailbox data after a      */
/*   read of write access to a mailbox        */
/*   mb: mailbox number (0..2)                */
/*   msg: mailbox data                        */
/* ****************************************** */
int ilmailbox(int mb, unsigned char *msg)
{
  int n;

  mb &= 3; /* security */

  /* test if msg received */
  n=msg[3];
  if (n&0x80) {
    if (state[mb]!=MBRESET) {
      state[mb]=MBRESET;
      reset_mb(mb);
    }
  }
  else {
    if (state[mb]==MBRESET)
      state[mb]=MBIDLE;
  }
  if (n&0x10) {
    if (n&0x8) send3bytes(mb,msg);
    else if (n&0x4) send1byte(mb,msg[0]);
    else execmd(mb,msg);
    /* clear msg available bit */
    msg[3] &= 0xef;
  }

  if (state[mb]==ACT_LISTEN) {
    n=msg[4];
    if ((n&01)==0)
      do_listen(mb, msg, lastframe);
  }

  if (state[mb]==END_LISTEN) {
    n=msg[4];
    if ((n&01)==0)
      end_listen(mb, msg);
  }

  if (fmanual) msg[4] |= 0x40; else msg[4] &= ~0x40;     /* 24/06/12 */

  return(0);
}



